@@ -1,34 +1,32 @@
 /*
- * eval.c - gawk bytecode interpreter 
+ * eval.c - gawk bytecode interpreter
  */
 
-/* 
- * Copyright (C) 1986, 1988, 1989, 1991-2013 the Free Software Foundation, Inc.
- * 
+/*
+ * Copyright (C) 1986, 1988, 1989, 1991-2019, 2021, 2022,
+ * the Free Software Foundation, Inc.
+ *
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
- * 
+ *
  * GAWK is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 3 of the License, or
  * (at your option) any later version.
- * 
+ *
  * GAWK is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
  */
 
 #include "awk.h"
+#include <math.h>
 
-extern void after_beginfile(IOBUF **curfile);
-extern double pow(double x, double y);
-extern double modf(double x, double *yp);
-extern double fmod(double x, double y);
 NODE **fcall_list = NULL;
 long fcall_count = 0;
 int currule = 0;
@@ -54,7 +52,7 @@ static NODE *node_Boolean[2];
 #ifdef C
 #undef C
 #endif
-#define C(c) ((char)c)  
+#define C(c) ((char)c)
 /*
  * This table is used by the regexp routines to do case independent
  * matching. Basically, every ascii character maps to itself, except
@@ -105,6 +103,12 @@ char casetable[] = {
 	'\170', '\171', '\172', '\173', '\174', '\175', '\176', '\177',
 
 	/* Latin 1: */
+	/*
+	 * 4/2019: This is now overridden; in single byte locales
+	 * we call load_casetable from main and it fills in the values
+	 * based on the current locale. In particular, we want LC_ALL=C
+	 * to work correctly for values >= 0200.
+	 */
 	C('\200'), C('\201'), C('\202'), C('\203'), C('\204'), C('\205'), C('\206'), C('\207'),
 	C('\210'), C('\211'), C('\212'), C('\213'), C('\214'), C('\215'), C('\216'), C('\217'),
 	C('\220'), C('\221'), C('\222'), C('\223'), C('\224'), C('\225'), C('\226'), C('\227'),
@@ -122,7 +126,7 @@ char casetable[] = {
 	C('\360'), C('\361'), C('\362'), C('\363'), C('\364'), C('\365'), C('\366'), C('\367'),
 	C('\370'), C('\371'), C('\372'), C('\373'), C('\374'), C('\375'), C('\376'), C('\377'),
 };
-#elif 'a' == 0x81 /* it's EBCDIC */
+#elif defined(USE_EBCDIC)
 char casetable[] = {
  /*00  NU    SH    SX    EX    PF    HT    LC    DL */
       0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
@@ -202,23 +206,20 @@ load_casetable(void)
 {
 #if defined(LC_CTYPE)
 	int i;
-	char *cp;
 	static bool loaded = false;
 
 	if (loaded || do_traditional)
 		return;
 
 	loaded = true;
-	cp = setlocale(LC_CTYPE, NULL);
 
-	/* this is not per standard, but it's pretty safe */
-	if (cp == NULL || strcmp(cp, "C") == 0 || strcmp(cp, "POSIX") == 0)
-		return;
-
-#ifndef ZOS_USS
+#ifndef USE_EBCDIC
+	/* use of isalpha is ok here (see is_alpha in awkgram.y) */
 	for (i = 0200; i <= 0377; i++) {
 		if (isalpha(i) && islower(i) && i != toupper(i))
 			casetable[i] = toupper(i);
+		else
+			casetable[i] = i;
 	}
 #endif
 #endif
@@ -237,10 +238,11 @@ static const char *const nodetypes[] = {
 	"Node_var",
 	"Node_var_array",
 	"Node_var_new",
+	"Node_elem_new",
 	"Node_param_list",
 	"Node_func",
 	"Node_ext_func",
-	"Node_old_ext_func",
+	"Node_builtin_func",
 	"Node_array_ref",
 	"Node_array_tree",
 	"Node_array_leaf",
@@ -285,6 +287,7 @@ static struct optypetab {
 	{ "Op_postincrement", "++" },
 	{ "Op_postdecrement", "--" },
 	{ "Op_unary_minus", "-" },
+	{ "Op_unary_plus", "+" },
 	{ "Op_field_spec", "$" },
 	{ "Op_not", "! " },
 	{ "Op_assign", " = " },
@@ -311,7 +314,7 @@ static struct optypetab {
 	{ "Op_match", " ~ " },
 	{ "Op_match_rec", NULL },
 	{ "Op_nomatch", " !~ " },
-	{ "Op_rule", NULL }, 
+	{ "Op_rule", NULL },
 	{ "Op_K_case", "case" },
 	{ "Op_K_default", "default" },
 	{ "Op_K_break", "break" },
@@ -322,20 +325,22 @@ static struct optypetab {
 	{ "Op_K_next", "next" },
 	{ "Op_K_exit", "exit" },
 	{ "Op_K_return", "return" },
+	{ "Op_K_return_from_eval", "return" },
 	{ "Op_K_delete", "delete" },
 	{ "Op_K_delete_loop", NULL },
 	{ "Op_K_getline_redir", "getline" },
 	{ "Op_K_getline", "getline" },
 	{ "Op_K_nextfile", "nextfile" },
+	{ "Op_K_namespace", "@namespace" },
 	{ "Op_builtin", NULL },
 	{ "Op_sub_builtin", NULL },
 	{ "Op_ext_builtin", NULL },
-	{ "Op_old_ext_builtin", NULL },	/* temporary */
 	{ "Op_in_array", " in " },
 	{ "Op_func_call", NULL },
 	{ "Op_indirect_func_call", NULL },
 	{ "Op_push", NULL },
 	{ "Op_push_arg", NULL },
+	{ "Op_push_arg_untyped", NULL },
 	{ "Op_push_i", NULL },
 	{ "Op_push_re", NULL },
 	{ "Op_push_array", NULL },
@@ -360,9 +365,11 @@ static struct optypetab {
 	{ "Op_after_beginfile", NULL },
 	{ "Op_after_endfile", NULL },
 	{ "Op_func", NULL },
+	{ "Op_comment", NULL },
 	{ "Op_exec_count", NULL },
 	{ "Op_breakpoint", NULL },
 	{ "Op_lint", NULL },
+	{ "Op_lint_plus", NULL },
 	{ "Op_atexit", NULL },
 	{ "Op_stop", NULL },
 	{ "Op_token", NULL },
@@ -377,6 +384,7 @@ static struct optypetab {
 	{ "Op_K_else", "else" },
 	{ "Op_K_function", "function" },
 	{ "Op_cond_exp", NULL },
+	{ "Op_parens", NULL },
 	{ "Op_final --- this should never appear", NULL },
 	{ NULL, NULL },
 };
@@ -395,7 +403,7 @@ nodetype2str(NODETYPE type)
 	return buf;
 }
 
-/* opcode2str --- convert a opcode type into a printable value */
+/* opcode2str --- convert an opcode type into a printable value */
 
 const char *
 opcode2str(OPCODE op)
@@ -406,6 +414,8 @@ opcode2str(OPCODE op)
 	return NULL;
 }
 
+/* op2str --- convert an opcode type to corresponding operator or keyword */
+
 const char *
 op2str(OPCODE op)
 {
@@ -432,17 +442,21 @@ flags2str(int flagval)
 		{ STRCUR, "STRCUR" },
 		{ NUMCUR, "NUMCUR" },
 		{ NUMBER, "NUMBER" },
-		{ MAYBE_NUM, "MAYBE_NUM" },
-		{ FIELD, "FIELD" },
+		{ USER_INPUT, "USER_INPUT" },
+		{ BOOLVAL, "BOOL" },
 		{ INTLSTR, "INTLSTR" },
 		{ NUMINT, "NUMINT" },
 		{ INTIND, "INTIND" },
 		{ WSTRCUR, "WSTRCUR" },
 		{ MPFN,	"MPFN" },
 		{ MPZN,	"MPZN" },
+		{ NO_EXT_SET, "NO_EXT_SET" },
+		{ NULL_FIELD, "NULL_FIELD" },
 		{ ARRAYMAXED, "ARRAYMAXED" },
 		{ HALFHAT, "HALFHAT" },
 		{ XARRAY, "XARRAY" },
+		{ NUMCONSTSTR, "NUMCONSTSTR" },
+		{ REGEX, "REGEX" },
 		{ 0,	NULL },
 	};
 
@@ -490,79 +504,79 @@ genflags2str(int flagval, const struct flagtab *tab)
 static int
 posix_compare(NODE *s1, NODE *s2)
 {
-	int ret = 0;
-	char save1, save2;
-	size_t l = 0;
-
-	save1 = s1->stptr[s1->stlen];
-	s1->stptr[s1->stlen] = '\0';
-
-	save2 = s2->stptr[s2->stlen];
-	s2->stptr[s2->stlen] = '\0';
+	int ret;
 
 	if (gawk_mb_cur_max == 1) {
-		if (strlen(s1->stptr) == s1->stlen && strlen(s2->stptr) == s2->stlen)
-			ret = strcoll(s1->stptr, s2->stptr);
-		else {
-			char b1[2], b2[2];
-			char *p1, *p2;
-			size_t i;
+		char save1, save2;
+		const char *p1, *p2;
 
-			if (s1->stlen < s2->stlen)
-				l = s1->stlen;
-			else
-				l = s2->stlen;
+		save1 = s1->stptr[s1->stlen];
+		s1->stptr[s1->stlen] = '\0';
 
-			b1[1] = b2[1] = '\0';
-			for (i = ret = 0, p1 = s1->stptr, p2 = s2->stptr;
-			     ret == 0 && i < l;
-			     p1++, p2++) {
-				b1[0] = *p1;
-				b2[0] = *p2;
-				ret = strcoll(b1, b2);
+		save2 = s2->stptr[s2->stlen];
+		s2->stptr[s2->stlen] = '\0';
+
+		p1 = s1->stptr;
+		p2 = s2->stptr;
+
+		for (;;) {
+			size_t len;
+
+			ret = strcoll(p1, p2);
+			if (ret != 0)
+				break;
+
+			len = strlen(p1);
+			p1 += len + 1;
+			p2 += len + 1;
+
+			if (p1 == s1->stptr + s1->stlen + 1) {
+				if (p2 != s2->stptr + s2->stlen + 1)
+					ret = -1;
+				break;
+			}
+			if (p2 == s2->stptr + s2->stlen + 1) {
+				ret = 1;
+				break;
 			}
 		}
-		/*
-		 * Either worked through the strings or ret != 0.
-		 * In either case, ret will be the right thing to return.
-		 */
+
+		s1->stptr[s1->stlen] = save1;
+		s2->stptr[s2->stlen] = save2;
 	}
-#if MBS_SUPPORT
 	else {
 		/* Similar logic, using wide characters */
+		const wchar_t *p1, *p2;
+
 		(void) force_wstring(s1);
 		(void) force_wstring(s2);
 
-		if (wcslen(s1->wstptr) == s1->wstlen && wcslen(s2->wstptr) == s2->wstlen)
-			ret = wcscoll(s1->wstptr, s2->wstptr);
-		else {
-			wchar_t b1[2], b2[2];
-			wchar_t *p1, *p2;
-			size_t i;
+		p1 = s1->wstptr;
+		p2 = s2->wstptr;
 
-			if (s1->wstlen < s2->wstlen)
-				l = s1->wstlen;
-			else
-				l = s2->wstlen;
+		for (;;) {
+			size_t len;
 
-			b1[1] = b2[1] = L'\0';
-			for (i = ret = 0, p1 = s1->wstptr, p2 = s2->wstptr;
-			     ret == 0 && i < l;
-			     p1++, p2++) {
-				b1[0] = *p1;
-				b2[0] = *p2;
-				ret = wcscoll(b1, b2);
+			ret = wcscoll(p1, p2);
+			if (ret != 0)
+				break;
+
+			len = wcslen(p1);
+			p1 += len + 1;
+			p2 += len + 1;
+
+			if (p1 == s1->wstptr + s1->wstlen + 1) {
+				if (p2 != s2->wstptr + s2->wstlen + 1)
+					ret = -1;
+				break;
+			}
+			if (p2 == s2->wstptr + s2->wstlen + 1) {
+				ret = 1;
+				break;
 			}
 		}
-		/*
-		 * Either worked through the strings or ret != 0.
-		 * In either case, ret will be the right thing to return.
-		 */
 	}
-#endif
 
-	s1->stptr[s1->stlen] = save1;
-	s2->stptr[s2->stlen] = save2;
 	return ret;
 }
 
@@ -570,7 +584,7 @@ posix_compare(NODE *s1, NODE *s2)
 /* cmp_nodes --- compare two nodes, returning negative, 0, positive */
 
 int
-cmp_nodes(NODE *t1, NODE *t2)
+cmp_nodes(NODE *t1, NODE *t2, bool use_strcmp)
 {
 	int ret = 0;
 	size_t len1, len2;
@@ -579,14 +593,8 @@ cmp_nodes(NODE *t1, NODE *t2)
 	if (t1 == t2)
 		return 0;
 
-	if ((t1->flags & MAYBE_NUM) != 0)
-		(void) force_number(t1);
-	if ((t2->flags & MAYBE_NUM) != 0)
-		(void) force_number(t2);
-	if ((t1->flags & INTIND) != 0)
-		t1 = force_string(t1);
-	if ((t2->flags & INTIND) != 0)
-		t2 = force_string(t2);
+	(void) fixtype(t1);
+	(void) fixtype(t2);
 
 	if ((t1->flags & NUMBER) != 0 && (t2->flags & NUMBER) != 0)
 		return cmp_numbers(t1, t2);
@@ -599,23 +607,28 @@ cmp_nodes(NODE *t1, NODE *t2)
 	if (len1 == 0 || len2 == 0)
 		return ldiff;
 
-	if (do_posix)
+	if (do_posix && ! use_strcmp)
 		return posix_compare(t1, t2);
 
 	l = (ldiff <= 0 ? len1 : len2);
 	if (IGNORECASE) {
 		const unsigned char *cp1 = (const unsigned char *) t1->stptr;
 		const unsigned char *cp2 = (const unsigned char *) t2->stptr;
+		char save1 = t1->stptr[t1->stlen];
+		char save2 = t2->stptr[t2->stlen];
+
 
-#if MBS_SUPPORT
 		if (gawk_mb_cur_max > 1) {
+			t1->stptr[t1->stlen] = t2->stptr[t2->stlen] = '\0';
 			ret = strncasecmpmbs((const unsigned char *) cp1,
 					     (const unsigned char *) cp2, l);
-		} else
-#endif
-		/* Could use tolower() here; see discussion above. */
-		for (ret = 0; l-- > 0 && ret == 0; cp1++, cp2++)
-			ret = casetable[*cp1] - casetable[*cp2];
+			t1->stptr[t1->stlen] = save1;
+			t2->stptr[t2->stlen] = save2;
+		} else {
+			/* Could use tolower() here; see discussion above. */
+			for (ret = 0; l-- > 0 && ret == 0; cp1++, cp2++)
+				ret = casetable[*cp1] - casetable[*cp2];
+		}
 	} else
 		ret = memcmp(t1->stptr, t2->stptr, l);
 
@@ -642,7 +655,7 @@ push_frame(NODE *f)
 	}
 
 	if (fcall_count > 1)
-		memmove(fcall_list + 2, fcall_list + 1, (fcall_count - 1) * sizeof(NODE *)); 
+		memmove(fcall_list + 2, fcall_list + 1, (fcall_count - 1) * sizeof(NODE *));
 	fcall_list[1] = f;
 }
 
@@ -653,7 +666,7 @@ static void
 pop_frame()
 {
 	if (fcall_count > 1)
-		memmove(fcall_list + 1, fcall_list + 2, (fcall_count - 1) * sizeof(NODE *)); 
+		memmove(fcall_list + 1, fcall_list + 2, (fcall_count - 1) * sizeof(NODE *));
 	fcall_count--;
 	assert(fcall_count >= 0);
 	if (do_debug)
@@ -667,7 +680,7 @@ void
 dump_fcall_stack(FILE *fp)
 {
 	NODE *f, *func;
-	long i = 0, j, k = 0;
+	long i = 0, k = 0;
 
 	if (fcall_count == 0)
 		return;
@@ -675,15 +688,13 @@ dump_fcall_stack(FILE *fp)
 
 	/* current frame */
 	func = frame_ptr->func_node;
-	for (j = 0; j <= frame_ptr->num_tail_calls; j++)
-		fprintf(fp, "\t# %3ld. %s\n", k++, func->vname);
+	fprintf(fp, "\t# %3ld. %s\n", k++, func->vname);
 
 	/* outer frames except main */
 	for (i = 1; i < fcall_count; i++) {
 		f = fcall_list[i];
 		func = f->func_node;
-		for (j = 0; j <= f->num_tail_calls; j++)
-			fprintf(fp, "\t# %3ld. %s\n", k++, func->vname);
+		fprintf(fp, "\t# %3ld. %s\n", k++, func->vname);
 	}
 
 	fprintf(fp, "\t# %3ld. -- main --\n", k);
@@ -696,28 +707,16 @@ void
 set_IGNORECASE()
 {
 	static bool warned = false;
-	NODE *n = IGNORECASE_node->var_value;
 
-	if ((do_lint || do_traditional) && ! warned) {
+	if ((do_lint_extensions || do_traditional) && ! warned) {
 		warned = true;
 		lintwarn(_("`IGNORECASE' is a gawk extension"));
 	}
-	load_casetable();
+
 	if (do_traditional)
 		IGNORECASE = false;
-	else if ((n->flags & (STRING|STRCUR)) != 0) {
-		if ((n->flags & MAYBE_NUM) == 0) {
-			(void) force_string(n);
-			IGNORECASE = (n->stlen > 0);
-		} else {
-			(void) force_number(n);
-			IGNORECASE = ! iszero(n);
-		}
-	} else if ((n->flags & (NUMCUR|NUMBER)) != 0)
-		IGNORECASE = ! iszero(n);
-	else
-		IGNORECASE = false;		/* shouldn't happen */
-                 
+   	else
+		IGNORECASE = boolval(IGNORECASE_node->var_value);
 	set_RS();	/* set_RS() calls set_FS() if need be, for us */
 }
 
@@ -728,16 +727,15 @@ set_BINMODE()
 {
 	static bool warned = false;
 	char *p;
-	NODE *v = BINMODE_node->var_value;
+	NODE *v = fixtype(BINMODE_node->var_value);
 
-	if ((do_lint || do_traditional) && ! warned) {
+	if ((do_lint_extensions || do_traditional) && ! warned) {
 		warned = true;
 		lintwarn(_("`BINMODE' is a gawk extension"));
 	}
 	if (do_traditional)
 		BINMODE = TEXT_TRANSLATE;
 	else if ((v->flags & NUMBER) != 0) {
-		(void) force_number(v);
 		BINMODE = get_number_si(v);
 		/* Make sure the value is rational. */
 		if (BINMODE < TEXT_TRANSLATE)
@@ -802,9 +800,35 @@ set_BINMODE()
 void
 set_OFS()
 {
+	static bool first = true;
+	size_t new_ofs_len;
+
+	if (first)	/* true when called from init_vars() in main() */
+		first = false;
+	else {
+		/* rebuild $0 using OFS that was current when $0 changed */
+		if (! field0_valid) {
+			get_field(UNLIMITED - 1, NULL);
+			rebuild_record();
+		}
+	}
+
+	/*
+	 * Save OFS value for use in building record and in printing.
+	 * Can't just have OFS point into the OFS_node since it's
+	 * already updated when we come into this routine, and we need
+	 * the old value to rebuild the record (see above).
+	 */
 	OFS_node->var_value = force_string(OFS_node->var_value);
-	OFS = OFS_node->var_value->stptr;
-	OFSlen = OFS_node->var_value->stlen;
+	new_ofs_len = OFS_node->var_value->stlen;
+
+	if (OFS == NULL)
+		emalloc(OFS, char *, new_ofs_len + 1, "set_OFS");
+	else if (OFSlen < new_ofs_len)
+		erealloc(OFS, char *, new_ofs_len + 1, "set_OFS");
+
+	memcpy(OFS, OFS_node->var_value->stptr, OFS_node->var_value->stlen);
+	OFSlen = new_ofs_len;
 	OFS[OFSlen] = '\0';
 }
 
@@ -816,7 +840,6 @@ set_ORS()
 	ORS_node->var_value = force_string(ORS_node->var_value);
 	ORS = ORS_node->var_value->stptr;
 	ORSlen = ORS_node->var_value->stlen;
-	ORS[ORSlen] = '\0';
 }
 
 /* fmt_ok --- is the conversion format a valid one? */
@@ -842,6 +865,8 @@ fmt_ok(NODE *n)
 	static const char flags[] = " +-#";
 #endif
 
+	// We rely on the caller to zero-terminate n->stptr.
+
 	if (*p++ != '%')
 		return 0;
 	while (*p && strchr(flags, *p) != NULL)	/* flags */
@@ -869,23 +894,30 @@ fmt_index(NODE *n)
 	int ix = 0;
 	static int fmt_num = 4;
 	static int fmt_hiwater = 0;
+	char save;
 
 	if (fmt_list == NULL)
 		emalloc(fmt_list, NODE **, fmt_num*sizeof(*fmt_list), "fmt_index");
 	n = force_string(n);
+
+	save = n->stptr[n->stlen];
+	n->stptr[n->stlen] = '\0';
+
 	while (ix < fmt_hiwater) {
-		if (cmp_nodes(fmt_list[ix], n) == 0)
+		if (cmp_nodes(fmt_list[ix], n, true) == 0)
 			return ix;
 		ix++;
 	}
+
 	/* not found */
-	n->stptr[n->stlen] = '\0';
 	if (do_lint && ! fmt_ok(n))
 		lintwarn(_("bad `%sFMT' specification `%s'"),
 			    n == CONVFMT_node->var_value ? "CONV"
 			  : n == OFMT_node->var_value ? "O"
 			  : "", n->stptr);
 
+	n->stptr[n->stlen] = save;
+
 	if (fmt_hiwater >= fmt_num) {
 		fmt_num *= 2;
 		erealloc(fmt_list, NODE **, fmt_num * sizeof(*fmt_list), "fmt_index");
@@ -919,49 +951,33 @@ set_LINT()
 {
 #ifndef NO_LINT
 	int old_lint = do_lint;
-	NODE *n = LINT_node->var_value;
+	NODE *n = fixtype(LINT_node->var_value);
 
-	if ((n->flags & (STRING|STRCUR)) != 0) {
-		if ((n->flags & MAYBE_NUM) == 0) {
-			const char *lintval;
-			size_t lintlen;
+	/* start with clean defaults */
+	lintfunc = r_warning;
+	do_flags &= ~(DO_LINT_ALL|DO_LINT_INVALID);
 
-			n = force_string(LINT_node->var_value);
-			lintval = n->stptr;
-			lintlen = n->stlen;
-			if (lintlen > 0) {
+	if ((n->flags & STRING) != 0) {
+		const char *lintval;
+		size_t lintlen;
+
+		lintval = n->stptr;
+		lintlen = n->stlen;
+		if (lintlen > 0) {
+			if (lintlen == 7 && strncmp(lintval, "invalid", 7) == 0)
+				do_flags |= DO_LINT_INVALID;
+			else if (lintlen == 6 && strncmp(lintval, "no-ext", 6) == 0)
+				do_flags &= ~DO_LINT_EXTENSIONS;
+			else {
 				do_flags |= DO_LINT_ALL;
 				if (lintlen == 5 && strncmp(lintval, "fatal", 5) == 0)
 					lintfunc = r_fatal;
-				else if (lintlen == 7 && strncmp(lintval, "invalid", 7) == 0) {
-					do_flags &= ~ DO_LINT_ALL;
-					do_flags |= DO_LINT_INVALID;
-				} else
-					lintfunc = warning;
-			} else {
-				do_flags &= ~(DO_LINT_ALL|DO_LINT_INVALID);
-				lintfunc = warning;
 			}
-		} else {
-			(void) force_number(n);
-			if (! iszero(n))
-				do_flags |= DO_LINT_ALL;
-			else
-				do_flags &= ~(DO_LINT_ALL|DO_LINT_INVALID);
-			lintfunc = warning;
 		}
-	} else if ((n->flags & (NUMCUR|NUMBER)) != 0) {
-		(void) force_number(n);
-		if (! iszero(n))
+	} else {
+		if (! is_zero(n))
 			do_flags |= DO_LINT_ALL;
-		else
-			do_flags &= ~(DO_LINT_ALL|DO_LINT_INVALID);
-		lintfunc = warning;
-	} else
-		do_flags &= ~(DO_LINT_ALL|DO_LINT_INVALID);	/* shouldn't happen */
-
-	if (! do_lint)
-		lintfunc = warning;
+	}
 
 	/* explicitly use warning() here, in case lintfunc == r_fatal */
 	if (old_lint != do_lint && old_lint && ! do_lint)
@@ -977,13 +993,10 @@ set_LINT()
 void
 set_TEXTDOMAIN()
 {
-	int len;
 	NODE *tmp;
 
 	tmp = TEXTDOMAIN_node->var_value = force_string(TEXTDOMAIN_node->var_value);
 	TEXTDOMAIN = tmp->stptr;
-	len = tmp->stlen;
-	TEXTDOMAIN[len] = '\0';
 	/*
 	 * Note: don't call textdomain(); this value is for
 	 * the awk program, not for gawk itself.
@@ -995,8 +1008,9 @@ set_TEXTDOMAIN()
 void
 update_ERRNO_int(int errcode)
 {
-	char *cp;
+	const char *cp;
 
+	update_PROCINFO_num("errno", errcode);
 	if (errcode) {
 		cp = strerror(errcode);
 		cp = gettext(cp);
@@ -1011,8 +1025,25 @@ update_ERRNO_int(int errcode)
 void
 update_ERRNO_string(const char *string)
 {
+	update_PROCINFO_num("errno", 0);
 	unref(ERRNO_node->var_value);
-	ERRNO_node->var_value = make_string(string, strlen(string));
+	size_t len = strlen(string);
+#if defined(USE_EBCDIC) && defined(ELIDE_IBM_ERROR_CODE)
+	// skip over leading IBM error code
+	// N.B. This code is untested
+	if (isupper(string[0]) && isupper(string[1])) {
+		while (*string && *string != ' ')
+			string++;
+
+		while (*string && *string == ' ')
+			string++;
+
+		len = strlen(string);
+		if (string[len-1] == '.')
+			len--;	// remove the final '.'
+	}
+#endif
+	ERRNO_node->var_value = make_string(string, len);
 }
 
 /* unset_ERRNO --- eliminate the value of ERRNO */
@@ -1020,6 +1051,7 @@ update_ERRNO_string(const char *string)
 void
 unset_ERRNO(void)
 {
+	update_PROCINFO_num("errno", 0);
 	unref(ERRNO_node->var_value);
 	ERRNO_node->var_value = dupnode(Nnull_string);
 }
@@ -1079,7 +1111,7 @@ STACK_ITEM *stack_bottom;
 STACK_ITEM *stack_top;
 static unsigned long STACK_SIZE = 256;    /* initial size of stack */
 int max_args = 0;       /* maximum # of arguments to printf, print, sprintf,
-                         * or # of array subscripts, or adjacent strings     
+                         * or # of array subscripts, or adjacent strings
                          * to be concatenated.
                          */
 NODE **args_array = NULL;
@@ -1126,7 +1158,7 @@ r_get_lhs(NODE *n, bool reference)
 					array_vname(n));
 		if (n->orig_array->type != Node_var) {
 			n->orig_array->type = Node_var;
-			n->orig_array->var_value = Nnull_string;
+			n->orig_array->var_value = dupnode(Nnull_string);
 		}
 		/* fall through */
 	case Node_var_new:
@@ -1134,11 +1166,19 @@ r_get_lhs(NODE *n, bool reference)
 		n->var_value = dupnode(Nnull_string);
 		break;
 
+	case Node_elem_new:
+		efree(n->stptr);
+		n->stptr = NULL;
+		n->stlen = 0;
+		n->type = Node_var;
+		n->var_value = dupnode(Nnull_string);
+		break;
+
 	case Node_var:
 		break;
 
 	default:
-		cant_happen();
+		cant_happen("unexpected variable type %s", nodetype2str(n->type));
 	}
 
 	if (do_lint && reference && var_uninitialized(n))
@@ -1151,8 +1191,8 @@ r_get_lhs(NODE *n, bool reference)
 
 
 /* r_get_field --- get the address of a field node */
- 
-static inline NODE **
+
+NODE **
 r_get_field(NODE *n, Func_ptr *assign, bool reference)
 {
 	long field_num;
@@ -1161,7 +1201,7 @@ r_get_field(NODE *n, Func_ptr *assign, bool reference)
 	if (assign)
 		*assign = NULL;
 	if (do_lint) {
-		if ((n->flags & NUMBER) == 0) {
+		if ((fixtype(n)->flags & NUMBER) == 0) {
 			lintwarn(_("attempt to field reference from non-numeric value"));
 			if (n->stlen == 0)
 				lintwarn(_("attempt to field reference from null string"));
@@ -1180,7 +1220,7 @@ r_get_field(NODE *n, Func_ptr *assign, bool reference)
 			*assign = reset_record;
 	} else
 		lhs = get_field(field_num, assign);
-	if (do_lint && reference && (*lhs == Null_field || *lhs == Nnull_string))
+	if (do_lint && reference && ((*lhs)->flags & NULL_FIELD) != 0)
 		lintwarn(_("reference to uninitialized field `$%ld'"),
 			      field_num);
 	return lhs;
@@ -1223,7 +1263,7 @@ calc_exp(AWKNUM x1, AWKNUM x2)
 }
 
 
-/* setup_frame --- setup new frame for function call */ 
+/* setup_frame --- setup new frame for function call */
 
 static INSTRUCTION *
 setup_frame(INSTRUCTION *pc)
@@ -1232,40 +1272,17 @@ setup_frame(INSTRUCTION *pc)
 	NODE *m, *f, *fp;
 	NODE **sp = NULL;
 	int pcount, arg_count, i, j;
-	bool tail_optimize = false;
 
 	f = pc->func_body;
 	pcount = f->param_cnt;
 	fp = f->fparms;
 	arg_count = (pc + 1)->expr_count;
 
-	/* tail recursion optimization */
-	tail_optimize =  ((pc + 1)->tail_call && do_optimize > 1
-				&& ! do_debug && ! do_profile);
-
-	if (tail_optimize) {
-		/* free local vars of calling frame */
-
-		NODE *func;
-		int n;
-
-		func = frame_ptr->func_node;
-		for (n = func->param_cnt, sp = frame_ptr->stack; n > 0; n--) {
-			r = *sp++;
-			if (r->type == Node_var)     /* local variable */
-				DEREF(r->var_value);
-			else if (r->type == Node_var_array)     /* local array */
-				assoc_clear(r);
-		}
-		sp = frame_ptr->stack;
-
-	} else if (pcount > 0) {
-		emalloc(sp, NODE **, pcount * sizeof(NODE *), "setup_frame");
-		memset(sp, 0, pcount * sizeof(NODE *));
+	if (pcount > 0) {
+		ezalloc(sp, NODE **, pcount * sizeof(NODE *), "setup_frame");
 	}
 
-
-	/* check for extra args */ 
+	/* check for extra args */
 	if (arg_count > pcount) {
 		warning(
 			_("function `%s' called with more arguments than declared"),
@@ -1278,13 +1295,9 @@ setup_frame(INSTRUCTION *pc)
 	}
 
 	for (i = 0, j = arg_count - 1; i < pcount; i++, j--) {
-		if (tail_optimize)
-			r = sp[i];
-		else {
-			getnode(r);
-			memset(r, 0, sizeof(NODE));
-			sp[i] = r;
-		}
+		getnode(r);
+		memset(r, 0, sizeof(NODE));
+		sp[i] = r;
 
 		if (i >= arg_count) {
 			/* local variable */
@@ -1297,10 +1310,17 @@ setup_frame(INSTRUCTION *pc)
 
 		if (m->type == Node_param_list)
 			m = GET_PARAM(m->param_cnt);
-			
+
+		/* $0 needs to be passed by value to a function */
+		if (m == fields_arr[0]) {
+			DEREF(m);
+			m = dupnode(m);
+		}
+
 		switch (m->type) {
 		case Node_var_new:
 		case Node_var_array:
+		case Node_elem_new:
 			r->type = Node_array_ref;
 			r->orig_array = r->prev_array = m;
 			break;
@@ -1325,19 +1345,21 @@ setup_frame(INSTRUCTION *pc)
 			r->var_value = m;
 			break;
 
+		case Node_func:
+		case Node_builtin_func:
+		case Node_ext_func:
+			r->type = Node_var;
+			r->var_value = make_string(m->vname, strlen(m->vname));
+			break;
+
 		default:
-			cant_happen();
+			cant_happen("unexpected parameter type %s", nodetype2str(m->type));
 		}
 		r->vname = fp[i].param;
 	}
 
 	stack_adj(-arg_count);	/* adjust stack pointer */
 
-	if (tail_optimize) {
-		frame_ptr->num_tail_calls++;
-		return f->code_ptr;
-	}
-
 	if (pc->opcode == Op_indirect_func_call) {
 		r = POP();	/* indirect var */
 		DEREF(r);
@@ -1353,11 +1375,10 @@ setup_frame(INSTRUCTION *pc)
 
 	/* setup new frame */
 	getnode(frame_ptr);
-	frame_ptr->type = Node_frame;	
+	frame_ptr->type = Node_frame;
 	frame_ptr->stack = sp;
 	frame_ptr->prev_frame_size = (stack_ptr - stack_bottom); /* size of the previous stack frame */
 	frame_ptr->func_node = f;
-	frame_ptr->num_tail_calls = 0;
 	frame_ptr->vname = NULL;
 	frame_ptr->reti = pc; /* on return execute pc->nexti */
 
@@ -1462,10 +1483,16 @@ unwind_stack(long n)
 			freenode(r);
 			break;
 		default:
-			if (in_main_context())
+			/*
+			 * Check `exiting' and don't produce an error for
+			 * cases like:
+			 *	func     _fn0() { exit }
+			 *	BEGIN { ARRAY[_fn0()] }
+			 */
+			if (in_main_context() && ! exiting)
 				fatal(_("unwind_stack: unexpected type `%s'"),
 						nodetype2str(r->type));
-			/* else 
+			/* else
 				* Node_var_array,
 				* Node_param_list,
 				* Node_var (e.g: trying to use scalar for array)
@@ -1479,7 +1506,7 @@ unwind_stack(long n)
 			break;
 	}
 	return cp;
-} 
+}
 
 
 /* pop_fcall --- pop off the innermost frame */
@@ -1489,7 +1516,7 @@ unwind_stack(long n)
 #define pop_stack()	(void) unwind_stack(0)
 
 
-static inline int
+static inline bool
 eval_condition(NODE *t)
 {
 	if (t == node_Boolean[false])
@@ -1498,39 +1525,118 @@ eval_condition(NODE *t)
 	if (t == node_Boolean[true])
 		return true;
 
-	if ((t->flags & MAYBE_NUM) != 0)
-		force_number(t);
-	else if ((t->flags & INTIND) != 0)
-		force_string(t);
-
-	if ((t->flags & NUMBER) != 0)
-		return ! iszero(t);
-
-	return (t->stlen != 0);
+	return boolval(t);
 }
 
+static bool cmp_doubles(const NODE *t1, const NODE *t2, scalar_cmp_t comparison_type);
+extern bool mpg_cmp_as_numbers(const NODE *t1, const NODE *t2, scalar_cmp_t comparison_type);
+
 /* cmp_scalars -- compare two nodes on the stack */
 
-static inline int
-cmp_scalars()
+static bool
+cmp_scalars(scalar_cmp_t comparison_type)
 {
 	NODE *t1, *t2;
 	int di;
+	bool ret;
 
 	t2 = POP_SCALAR();
 	t1 = TOP();
+
+	t1 = elem_new_to_scalar(t1);
+	t2 = elem_new_to_scalar(t2);
+
 	if (t1->type == Node_var_array) {
 		DEREF(t2);
 		fatal(_("attempt to use array `%s' in a scalar context"), array_vname(t1));
 	}
-	di = cmp_nodes(t1, t2);
+
+	if ((t1->flags & STRING) != 0 || (t2->flags & STRING) != 0) {
+		bool use_strcmp = (comparison_type == SCALAR_EQ || comparison_type == SCALAR_NEQ);
+		di = cmp_nodes(t1, t2, use_strcmp);
+
+		switch (comparison_type) {
+		case SCALAR_EQ:
+			ret = (di == 0);
+			break;
+		case SCALAR_NEQ:
+			ret = (di != 0);
+			break;
+		case SCALAR_LT:
+			ret = (di < 0);
+			break;
+		case SCALAR_LE:
+			ret = (di <= 0);
+			break;
+		case SCALAR_GT:
+			ret = (di > 0);
+			break;
+		case SCALAR_GE:
+			ret = (di >= 0);
+			break;
+		}
+	} else {
+		fixtype(t1);
+		fixtype(t2);
+
+#ifdef HAVE_MPFR
+		if (do_mpfr)
+			ret = mpg_cmp_as_numbers(t1, t2, comparison_type);
+		else
+#endif
+			ret = cmp_doubles(t1, t2, comparison_type);
+	}
+
 	DEREF(t1);
 	DEREF(t2);
-	return di;
+	return ret;
 }
 
+
+/* cmp_doubles --- compare two doubles */
+
+static bool
+cmp_doubles(const NODE *t1, const NODE *t2, scalar_cmp_t comparison_type)
+{
+	/*
+	 * This routine provides numeric comparisons that should work
+	 * the same as in C.  It should NOT be used for sorting.
+	 */
+
+	bool t1_nan = isnan(t1->numbr);
+	bool t2_nan = isnan(t2->numbr);
+	int ret;
+
+	if ((t1_nan || t2_nan) && comparison_type != SCALAR_NEQ)
+		return false;
+
+	switch (comparison_type) {
+	case SCALAR_EQ:
+		ret = (t1->numbr == t2->numbr);
+		break;
+	case SCALAR_NEQ:
+		ret = (t1->numbr != t2->numbr);
+		break;
+	case SCALAR_LT:
+		ret = (t1->numbr < t2->numbr);
+		break;
+	case SCALAR_LE:
+		ret = (t1->numbr <= t2->numbr);
+		break;
+	case SCALAR_GT:
+		ret = (t1->numbr > t2->numbr);
+		break;
+	case SCALAR_GE:
+		ret = (t1->numbr >= t2->numbr);
+		break;
+	}
+
+	return ret;
+}
+
+
 /* op_assign --- assignment operators excluding = */
- 
+
 static void
 op_assign(OPCODE op)
 {
@@ -1632,9 +1738,9 @@ POP_CODE()
 typedef struct exec_state {
 	struct exec_state *next;
 
-	INSTRUCTION *cptr;  /* either getline (Op_K_getline) or the 
+	INSTRUCTION *cptr;  /* either getline (Op_K_getline) or the
 	                     * implicit "open-file, read-record" loop (Op_newfile).
-	                     */ 
+	                     */
 
 	int rule;           /* rule for the INSTRUCTION */
 
@@ -1700,7 +1806,7 @@ register_exec_hook(Func_pre_exec preh, Func_post_exec posth)
 	/*
 	 * multiple post-exec hooks aren't supported. post-exec hook is mainly
 	 * for use by the debugger.
-	 */ 
+	 */
 
 	if (! preh || (post_execute && posth))
 		return false;
@@ -1728,7 +1834,7 @@ register_exec_hook(Func_pre_exec preh, Func_post_exec posth)
 }
 
 
-/* interpreter routine when not debugging */ 
+/* interpreter routine when not debugging */
 #include "interpret.h"
 
 /* interpreter routine with exec hook(s). Used when debugging and/or with MPFR. */
@@ -1756,7 +1862,6 @@ init_interpret()
 	frame_ptr->type = Node_frame;
 	frame_ptr->stack = NULL;
 	frame_ptr->func_node = NULL;	/* in main */
-	frame_ptr->num_tail_calls = 0;
 	frame_ptr->vname = NULL;
 
 	/* initialize true and false nodes */
@@ -1776,6 +1881,23 @@ init_interpret()
 	if (num_exec_hook > 0)
 		interpret = h_interpret;
 	else
-		interpret = r_interpret; 
+		interpret = r_interpret;
 }
 
+/* elem_new_to_scalar --- convert Node_elem_new to untyped scalar */
+
+NODE *
+elem_new_to_scalar(NODE *n)
+{
+	if (n->type != Node_elem_new)
+		return n;
+
+	if (n->valref > 1) {
+		unref(n);
+		return dupnode(Nnull_string);
+	}
+
+	n->type = Node_val;
+
+	return n;
+}
