@@ -3,7 +3,7 @@
  */
 
 /*
- * Copyright (C) 1986, 1988, 1989, 1991-2015, 2017-2020,
+ * Copyright (C) 1986, 1988, 1989, 1991-2015, 2017-2020, 2022,
  * the Free Software Foundation, Inc.
  *
  * This file is part of GAWK, the GNU implementation of the
@@ -50,10 +50,10 @@ NODE *symbol_table, *func_table;
 /* Use a flag to avoid a strcmp() call inside install() */
 static bool installing_specials = false;
 
-/* init_symbol_table --- make sure the symbol tables are initialized */
+/* init_the_tables --- deal with the tables for in memory use */
 
-void
-init_symbol_table()
+static void
+init_the_tables(void)
 {
 	getnode(global_table);
 	memset(global_table, '\0', sizeof(NODE));
@@ -65,11 +65,57 @@ init_symbol_table()
 
 	installing_specials = true;
 	func_table = install_symbol(estrdup("FUNCTAB", 7), Node_var_array);
-
 	symbol_table = install_symbol(estrdup("SYMTAB", 6), Node_var_array);
 	installing_specials = false;
 }
 
+/* init_symbol_table --- make sure the symbol tables are initialized */
+
+void
+init_symbol_table()
+{
+	if (! using_persistent_malloc) {
+		// normal case, initialize regularly, return
+		init_the_tables();
+		return;
+	}
+
+	// using persistent memory, get the root pointer
+	// which holds this struct:
+	struct root_pointers {
+		NODE *global_table;
+		NODE *func_table;
+	} *root_pointers = NULL;
+
+	root_pointers = (struct root_pointers *) pma_get_root();
+
+	if (root_pointers == NULL) {
+		// very first time!
+
+		// set up the tables
+		init_the_tables();
+
+		// save the pointers for the next time.
+		emalloc(root_pointers, struct root_pointers *, sizeof(struct root_pointers), "init_symbol_table");
+		root_pointers->global_table = global_table;
+		root_pointers->func_table = func_table;
+		pma_set_root(root_pointers);
+	} else {
+		// this is the next time, get the saved pointers and put them back in place
+		global_table = root_pointers->global_table;
+		func_table = root_pointers->func_table;
+
+		// still need to set this one up as usual
+		getnode(param_table);
+		memset(param_table, '\0', sizeof(NODE));
+		null_array(param_table);
+
+		// set the global variables
+		symbol_table = lookup("SYMTAB");
+		func_table = lookup("FUNCTAB");
+	}
+}
+
 /*
  * install_symbol:
  * Install a global name in the symbol table, even if it is already there.
@@ -479,6 +525,8 @@ print_vars(NODE **table, int (*print_func)(FILE *, const char *, ...), FILE *fp)
 			print_func(fp, "untyped variable\n");
 		else if (r->type == Node_var)
 			valinfo(r->var_value, print_func, fp);
+		else
+			cant_happen("unexpected node type: %s", nodetype2str(r->type));
 	}
 }
 
@@ -604,7 +652,10 @@ load_symbols()
 			    || r->type == Node_var
 			    || r->type == Node_var_array
 			    || r->type == Node_var_new) {
-				tmp = make_string(r->vname, strlen(r->vname));
+				if (strncmp(r->vname, "awk::", 5) == 0)
+					tmp = make_string(r->vname + 5, strlen(r->vname) - 5);
+				else
+					tmp = make_string(r->vname, strlen(r->vname));
 				aptr = assoc_lookup(sym_array, tmp);
 				unref(tmp);
 				unref(*aptr);
@@ -628,7 +679,7 @@ load_symbols()
 					*aptr = dupnode(untyped);
 					break;
 				default:
-					cant_happen();
+					cant_happen("unexpected node type %s", nodetype2str(r->type));
 					break;
 				}
 			}
@@ -641,6 +692,7 @@ load_symbols()
 	unref(scalar);
 	unref(untyped);
 	unref(array);
+	unref(built_in);
 }
 
 /* check_param_names --- make sure no parameter is the name of a function */
@@ -918,7 +970,7 @@ free_bc_internal(INSTRUCTION *cp)
 			unref(m);
 		break;
 	case Op_illegal:
-		cant_happen();
+		cant_happen("unexpected opcode %s", opcode2str(cp->opcode));
 	default:
 		break;
 	}
